// ulimit -v 16777216
// gcc -Wall -Wpedantic -Wextra -std=c17 -march=native -O3 memory-latency.c && perf stat -e cache-misses,cache-references ./a.out 30 28
// gcc -Wall -Wpedantic -Wextra -std=c17 -march=native -O3 memory-latency.c && ./a.out 30 28

#define _POSIX_C_SOURCE 199309L
#include <time.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define DATA_T uint8_t
#define SUM_T uint64_t
#define MAX_VAL 255
#define BUFFER_SIZE 102400


DATA_T* build_array(size_t length) {
    printf("Allocating %g MB (this should be many times larger than L3 cache)...\n", (double)length * sizeof(DATA_T) / 1024 / 1024);
    DATA_T *data = (DATA_T*)malloc(length * sizeof(DATA_T));
    if (data == NULL) {
        printf("Couldn't allocate.\n");
        return data;
    }
    
    printf("Filling array...\n");
    for (size_t i = 0; i < length; i++) {
        data[i] = rand() % MAX_VAL;
    }
    return data;
}


double cpu_clock_speed() {
    // Get the max CPU clock speed from lscpu. I'm not proud of this: it's incredibly fragile code.
    FILE *fp;
    char output[BUFFER_SIZE];
    __attribute__((unused)) size_t read;
    char *marker = "CPU max MHz:";
    char *start, *end;

    fp = popen("lscpu", "r");
    if (fp == NULL) {
        printf("Failed to run command\n" );
        return 0;
    }
    
    // read the lscpu output
    read = fread(output, BUFFER_SIZE, 1, fp);
    
    // find the start of the max MHz in the string
    start = strstr(output, marker);
    if ( start == NULL ) {
        printf("Failed to get max frequency\n" );
        return 0;
    }
    start += strlen(marker);
    while( isspace(*start) ) {
        start++;
    }
    
    // find the end of it, assuming it's there
    end = start;
    while( ! isspace(*end) ) {
        end++;
    }
    *end = '\0';
    return atof(start);
}


int test(size_t length, size_t reads) {
    struct timespec start, end;
    
    srand(time(NULL));
    DATA_T* data = build_array(length);
    if (data == NULL) {
        return 1;
    }
    
    // read random values from the array
    printf("Reading from array %g M times...\n", (double)reads / 1024 / 1024);
    SUM_T total = 0;
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start);
    for (size_t i = 0; i < reads; i++) {
        total += data[rand() % length];
    }
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &end);
    printf("Calculated %lu\n", total);
    double elapsed = (end.tv_sec - start.tv_sec)
                 + (end.tv_nsec - start.tv_nsec) / 1e9;
    printf("Elapsed: %g s\n", elapsed);
    printf("Per access: %g ns\n", elapsed / reads * 1e9);
    
    double freq = cpu_clock_speed();
    if ( freq != 0 ) {
        printf("         == %g cycles at max CPU frequency\n", elapsed / reads * 1e6 * freq);
    }

    free(data);
    return 0;
}


int main(int argc, char* argv[]) {
    size_t length, reads;
    if (argc < 2) {
        length = 1 << 26;
        reads = 1 << 26;
    } else if (argc < 3) {
        length = 1 << atol(argv[1]);
        reads = length;
    } else {
        length = 1 << atol(argv[1]);
        reads = 1 << atol(argv[2]);
    }
    return test(length, reads);
}
